/* config.c - Parse the TextConfig file for VGA configuration */

#include "externs.h"

struct modeline *modelist;
struct font *fontlist;
struct range *horiz_ranges, *vert_ranges;

static char *flaglist[]={"HSync","+HSync","-HSync", "VSync","+VSync","-VSync",
                         "Doublescan", "Interlace", "SyncOnGreen", 0};
static int flaglist_values[]={~NHSYNC,~NHSYNC,NHSYNC, ~NVSYNC,~NVSYNC,NVSYNC,
                              DOUBLESCAN, INTERLACE, CSYNC, 0};

char *modeline_flags[]={"-HSync", "-VSync", "Doublescan", "Interlace",
                        "SyncOnGreen", 0};

static char *directives[]={
  /* Monitor Range */
  "HorizSync", "VertRefresh",
  
  /* Default Settings */
  "DefaultMode", "DefaultRefresh", "DacSpeed", "Cursor", "Underline", "Blink",
  "IsoFont", "LineArt", "Palette", "BorderColor", "FontSize", "FontSelect",
  "FontProg", "ResetProg",

  /* Card-specific Config */
  "Option", "Chipset", "ClockChip", 0};

static char *option_flags[]={
  "LoadFont", "Override", "", "",

  "Slow_DRAM", "Med_DRAM", "Fast_DRAM", "XFast_DRAM",
  "Hibit_High", "Swap_Hibit", "S3_HSText", "Legend", "SPEA_Mercury",
  "Clockchip_x", "Topleft", "Stretch",
  0};

static char *chipset_list[]={
  "SVGA", "nVidia", "ATI", "Matrox", "S3", "Cirrus", "ET3000", "ET4000",
  "ET6000", "TVGA8900", "TVGA9000", "PVGA1", "WDC90C0X", "WDC90C1X",
  "WDC90C2X", "WDC90C3X", "ATI28800", "ATIMach32", "ATIMach64", "Headland",
  "ALI", "ALI2101", "OTI67", "OTI77", "OTI87", "SiS", "RealTek", "ARK",
  "NCR77C22E", "NCR77C32", "GVGA", "MX", "TGUI", "NeoMagic",
  0};

static char *clockchip_list[]={
  "ICD2061A", "ICS9161A", "ICS2595", "DCS2834", "ICS5300", "ICS5341",
  "ICS5342", "SC11412", "S3GenDAC", "S3_SDAC", "S3Trio", "S3Virge", "TI3025",
  "TI026", "CH8391", "STG1703", "IBM_RGB5xx", "ET6000", "Mach64", "TGUI9320",
  "TGUI9440", "Cyber938x", "ATT20C408", "MGA1064", "Laguna",
  0};


/* Get line from file removing leading \r\n */
static char *ftext(char *buf, int len, FILE *f)
{
  char *s;

  if(!fgets(buf, len, f))
    return NULL;

  s=buf+strlen(buf)-1;
  if(s >= buf && (*s == '\n' || *s == '\r'))
    *s--='\0';
  if(s >= buf && *s == '\r')
    *s='\0';
  return buf;
}

/* Separates string by whitespace */
static char *strspc(char **s)
{
  char *str=*s;

  if(!**s)
    return NULL;
  while(**s && !isspace(**s))
    (*s)++;
  while(isspace(**s))
    *(*s)++='\0';
  return str;
}


/* Read the config file, storing all values read in globals */
int read_config(char *filename)
{
  FILE *f;
  struct modeline *mode, *lastmode;
  struct font *font, *last;
  struct range *range;
  char buf[256], *r, *s, *t, *u;
  int i, val, val2, line=0;

  if(!(f=fopen(filename, "r")))
    return 0;

  if(verbose)
    printf("Loading textmode configuration from '%s'...\n", filename);

  while(ftext(buf, 256, f)) {
    line++;

    /* Skip leading whitespace */
    for(s=buf;isspace(*s);s++);

    /* Blank out all comments */
    for(r=s;*r;r++) {
      if(*r == '#') {
        *r='\0';
        break;
      }
      if(*r == '"') {
        while(*r++ && *r != '"');
        if(!*r)
          break;
      }
    }

    if(!*s)
      continue;  /* Blank line */

    /* Parse Modeline Directive */
    if(!strncasecmp(s, "Modeline", 8) && isspace(*(s+8))) {
      for(s+=8;isspace(*s);s++);
      if(*s != '"') {
        fprintf(stderr, "Warning: Invalid Modeline directive in %s line %d\n",
                filename, line);
        continue;
      }
    }

    /* Modelines begin with "Name" and contain at least 9 values following */
    if(*s == '"') {
      if(!(r=strchr(++s, '"'))) {
        fprintf(stderr, "Missing ending \" on modeline in %s line %d\n",
                filename, line);
        continue;
      }
      *r++='\0';
      if(!*s) {
        fprintf(stderr, "Blank modeline definition in %s line %d\n", filename,
                line);
        continue;
      }

      for(lastmode=NULL,mode=modelist;mode;mode=mode->next) {
        lastmode=mode;
        if(!mode->next)
          break;
      }
      mode=calloc(1, sizeof(struct modeline));
      mode->next=NULL;
      mode->width=9;
      mode->height=16;
      memcpy(mode->name=malloc(r-s), s, r-s);
      for(s=r;isspace(*s);s++);

      /* Read Pixel Clock (in MHz) */
      if(!(r=strspc(&s))) {
        fprintf(stderr, "Missing pixel clock in %s line %d\n", filename, line);
        free(mode);
        continue;
      }
      mode->clock=atof(r)*1000;  /* Convert to kHz */
      if(mode->clock < 10000) {
        fprintf(stderr, "Pixel clock too low (%d.%03d MHz) in %s line %d\n",
                mode->clock/1000, mode->clock%1000, filename, line);
        free(mode);
        continue;
      }

      /* Read in 8 timing values */
      for(i=0;i<8;i++) {
        if(!(r=strspc(&s))) {
          fprintf(stderr, "Missing modeline timing values in %s line %d\n",
                  filename, line);
          break;
        }
        mode->timing[i]=atoi(r);
      }
      if(i != 8) {
        free(mode);
        continue;
      }

      /* Read additional modeline flags */
      while((r=strspc(&s))) {
        for(i=0;flaglist[i];i++)
          if(!strcasecmp(flaglist[i], r)) {
            if(flaglist_values[i] & 0x80000000)
              mode->flags &= flaglist_values[i];
            else
              mode->flags |= flaglist_values[i];
            break;
          }
        if(flaglist[i])
          continue;

        /* Parse "Font" Option */
        if(!strcasecmp("Font", r) && (r=strspc(&s))) {
          if(!isdigit(*r)) {
            badfont:
            fprintf(stderr, "Warning: Bad font size '%s' in %s line %d\n", r,
                    filename, line);
            continue;
          }
          for(t=r;isdigit(*t);t++);
          if(!*t++)
            goto badfont;

          mode->width=atoi(r);
          mode->height=atoi(t);
          if((mode->width != 8 && mode->width != 9) ||
             mode->height < 1 || mode->height > 32)
            goto badfont;

          continue;
        }
        fprintf(stderr, "Warning: Unknown mode flag '%s' in %s line %d\n",
                r, filename, line);
      }

      /* Add modeline to end of linked list */
      if(lastmode)
        lastmode->next=mode;
      else
        modelist=mode;
      continue;
    }

    /* Parse Config Options */
    r=strspc(&s);
    for(i=0;directives[i];i++)
      if(!strcasecmp(directives[i], r))
        break;
    if(!directives[i]) {
      fprintf(stderr, "Unknown configuration directive '%s' in %s line %d\n",
              r, filename, line);
      continue;
    }

    for(;isspace(*s);s++);

    /* Remove double-quotes around arguments for directives such as DefaultMode
       and FontSelect. 'r' gets set to the text following the argument. */

    if(*s == '"') {
      if(i < 2) {
        fprintf(stderr, "Warning: Directive %s must not have arguments in "
                "quotes in %s line %d\n", directives[i], filename, line);
        continue;
      }
      for(r=++s;*r && *r != '"';r++);
      if(*r) {
        *r++='\0';
        for(;isspace(*r);r++);
      }
    } else {
      if(i == 13) {
        fprintf(stderr, "Warning: Directive %s requires an argument in quotes "
                "in %s line %d\n", directives[i], filename, line);
        continue;
      } if(!*s) {
        fprintf(stderr, "Directive %s requires an argument in %s line %d\n",
                directives[i], filename, line);
        continue;
      }
    }


    switch(i) {
    case 0:  /* Horizontal Ranges */
    case 1:  /* Vertical Ranges */
      while((r=strspc(&s))) {
        if(!isdigit(*r)) {
          badrange:
          fprintf(stderr, "Warning: Invalid monitor refresh rate '%s' in "
                  "%s line %d\n", r, filename, line);
          continue;
        }
        val=atof(r)*1000;
        for(;isdigit(*r);r++);
        if(*r) {
          if(*r++ != '-')
            goto badrange;
          val2=atof(r)*1000;
          if(val2 < val)
            goto badrange;
        } else
          val2=val;
        if((i == 0 && (val < 12000 || val > 150000 ||
                       val2 < 12000 || val2 > 150000)) ||
           (i == 1 && (val < 50000 || val > 250000 ||
                       val2 < 50000 || val2 > 250000)))
          goto badrange;

        /* Add range to list (with a variance of -0.5 to +0.5) */
        range=malloc(sizeof(struct range));
        range->next=(i == 0)?horiz_ranges:vert_ranges;
        range->lo=val-500;
        range->hi=val2+500;
        (i == 0)?horiz_ranges:vert_ranges=range;
      }
      break;
    case 2:  /* Default Mode */
      free(modeline);
      modeline=strdup(s);
      break;
    case 3:  /* Default Refresh */
      val=atof(s)*1000;
      if(val < 50000 || val > 250000)
        fprintf(stderr, "Warning: Invalid refresh rate '%d.%3d Hz' in %s "
                "line %d\n", val/1000, val%1000, filename, line);
      else
        refresh=val;
      break;
    case 4:  /* Dac Speed */
      val=atof(s)*1000+500;
      if(val > 250000) {
        fprintf(stderr, "Warning: Invalid DacSpeed pixel clock '%d.%3d MHz in "
                "%s line %d\n", val/1000, val%1000, filename, line);
        fprintf(stderr, "DacSpeed refers to the maximum clock for text modes "
          "and is much lower\nthan the rated value for graphics modes.\n");
      } else
        maxclk=val;
      break;
    case 5:  /* Cursor */
      if(!isdigit(*s)) {
        badcursor:
        fprintf(stderr, "Warning: Cursor range can only be from 0 to 31 in "
                "%s line %d\n", filename, line);
        break;
      }
      val=atoi(s);
      for(;isdigit(*s);s++);
      if(*s == '-') {
        s++;
        if(!isdigit(*s))
          goto badcursor;
        val2=atoi(s);
      } else if(*s) {
        goto badcursor;
      } else
        val2=val;
      if(val < 0 || val > 31 || val2 < 0 || val2 > 31)
        goto badcursor;

      cursor_start=val;
      cursor_end=val2;
      break;
    case 6:  /* Underline */
      val=atoi(s);
      if(val > 0 || val > 31)
        fprintf(stderr, "Warning: Underline can only be from 0 to 31 in "
                "%s line %d\n", filename, line);
      else
        underline=val;
      break;
    case 7:  /* Blink */
      if(!strcasecmp(s, "On"))
        blink=1;
      else if(!strcasecmp(s, "Off"))
        blink=0;
      else
        fprintf(stderr, "Warning: Blink can only be 'on' or 'off' in "
                "%s line %d\n", filename, line);
      break;
    case 8:  /* Iso Font (512-byte) */
      if(!strcasecmp(s, "On"))
        isofont=1;
      else if(!strcasecmp(s, "Off"))
        isofont=0;
      else
        fprintf(stderr, "Warning: IsoFont can only be 'on' or 'off' in "
                "%s line %d\n", filename, line);
      break;
    case 9:  /* Line Art */
      if(!strcasecmp(s, "On"))
        lineart=1;
      else if(!strcasecmp(s, "Off"))
        lineart=0;
      else
        fprintf(stderr, "Warning: LineArt can only be 'on' or 'off' in "
                "%s line %d\n", filename, line);
      break;
    case 10:  /* Palette 6/8-bit */
      val=atoi(s);
      if(val != 6 && val != 8)
        fprintf(stderr, "Warning: Palette can only be 6 or 8 bits wide in "
                "%s line %d\n", filename, line);
      else
        palette=(val == 8);
      break;
    case 11:  /* Border Color */
      val=atoi(s);
      if(val < 0 || val > 255)
        fprintf(stderr, "Warning: BorderColor can only be from 0 to 255 in "
                "%s line %d\n", filename, line);
      else
        border_color=val;
      break;
    case 12:  /* Default Font Size */
      if(!isdigit(*s)) {
        badfontsz:
        fprintf(stderr, "Warning: Invalid default font size '%s' in %s line %d\n",
                s, filename, line);
        break;
      }
      for(t=s;isdigit(*t);t++);
      if(!*t++)
        goto badfontsz;

      val=atoi(s);
      val2=atoi(t);
      if((val != 8 && val != 9) || val2 < 1 || val2 > 32)
        goto badfontsz;

      fontwidth=val;
      fontheight=val2;
      break;
    case 13:  /* Font Select */
      while((t=strspc(&r))) {
        if(!isdigit(*t)) {
          badfontsz2:
          fprintf(stderr, "Warning: Invalid font size '%s' in %s line %d\n",
                  t, filename, line);
          continue;
        }
        for(u=t;isdigit(*u);u++);
        if(!*u++) {  /* Format: 32 */
          val=0;
          val2=atoi(t);
        } else {     /* Format: 8x32 or 9x32 */
          val=atoi(t);
          val2=atoi(u);
          if(val != 8 && val != 9)
            goto badfontsz2;
        }
      
        if(val2 < 1 || val2 > 32)
          goto badfontsz2;

        /* Add new font selection to end of fontlist */
        for(last=NULL,font=fontlist;font;font=font->next) {
          last=font;
          if(!font->next)
            break;
        }
        font=malloc(sizeof(struct font));
        font->next=NULL;
        font->name=strdup(s);
        font->width=val;
        font->height=val2;
        if(last)
          last->next=font;
        else
          fontlist=font;
      }
      break;
    case 14:  /* Font Program */
      free(fontprog);
      fontprog=strdup(s);
      break;
    case 15:  /* Reset Program */
      free(resetprog);
      resetprog=strdup(s);
      break;
    case 16:  /* Option */
      for(i=0;option_flags[i];i++)
        if(!strcasecmp(option_flags[i], s))
          break;
      if(!option_flags[i]) {
        fprintf(stderr, "Warning: Unrecognized chipset option '%s' in "
                "%s line %d\n", s, filename, line);
        continue;
      }
      options |= (1 << i);
      break;
    case 17:  /* Chipset */
      if(chipset != -1) {
        fprintf(stderr, "Error: Multiple chipset definition in %s line %d\n",
                filename, line);
        exit(1);
      }
      for(i=0;chipset_list[i];i++)
        if(!strcasecmp(chipset_list[i], s))
          break;
      if(!chipset_list[i]) {
        fprintf(stderr, "Error: Unknown chipset '%s' in %s line %d\n", s,
                filename, line);
        exit(1);
      }
      chipset=i;
      break;
    case 18:  /* ClockChip */
      if(clockchip != -1) {
        fprintf(stderr, "Error: Multiple clockchip definition in %s line %d\n",
                filename, line);
        exit(1);
      }
      for(i=0;clockchip_list[i];i++)
        if(!strcasecmp(clockchip_list[i], s))
          break;
      if(!clockchip_list[i]) {
        fprintf(stderr, "Error: Unknown clockchip '%s' in %s line %d\n", s,
                filename, line);
        exit(1);
      }
      clockchip=i;
      break;
    }
  }

  fclose(f);
  return 1;
}
