/* ttyresize.c - Terminal resizing code for SVGATextMode */

#include "externs.h"
#include <sys/ioctl.h>
#include <termios.h>

#ifdef linux

/* Linux Port by Byron Stanoszek */

/* Check kernel version */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,0,0)
# error "SVGATextMode only supports Linux Kernels 2.0 and up."
#endif

#include <asm/page.h>   /* for PAGE_SIZE */
#include <linux/vt.h>   /* for VT_RESIZE */

// Resizes the Linux Console.
// Returns 0 on failure, or 1 if resized.
//
int resize_tty(int cols, int rows)
{
  char buf[16];
  struct vt_consize geom;
  int fd, i, max=0;

  /* Find number of open TTYs */
  if((fd=open("/dev/console", O_RDWR|O_NOCTTY)) == -1) {
    perror("/dev/console");
    return 0;
  }
  ioctl(fd, VT_OPENQRY, &max);
  close(fd);
  if(max < 2)
    max=64;

  /* Loop for all open TTYs */
  for(i=1;i < max;i++) {
    sprintf(buf, "/dev/tty%d", i);
    if((fd=open(buf, O_RDWR|O_NOCTTY)) == -1)
      continue;

    /* Resize TTY */
    geom.v_rows=rows;
    geom.v_cols=cols;
    geom.v_vlin=rows*fontheight;
    geom.v_clin=fontheight;
    geom.v_vcol=cols*fontwidth;
    geom.v_ccol=fontwidth;

    ioctl(fd, VT_RESIZEX, &geom);
    close(fd);
  }

  return 1;  /* Successful */
}

#elif defined(hurd)

/* Mach/HURD port by Arkadi E. Shishlov <arkadi@it.lv> */

#include <mach.h>
#include <hurd.h>
#include <device/device.h>

/* Mach/HURD ioctl() schemes are incompatible, use magic number here. */

#undef KDRESIZE
#define KDRESIZE 0x80086b05

struct kdvtsize {
  int lines;
  int cols;
};

// Resizes the GNU Hurd Console.
// Returns 0 on failure, 1 if resized, 2 if no resizing was necessary.
//
int resize_tty(int cols, int rows)
{
  mach_port_t dev, console_dev;
  kern_return_t err;
  struct kdvtsize sz;
  struct winsize winsize;

  if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &winsize)) {
    perror("Get Window Size");
    return 0;
  }

  if(winsize.ws_col == cols && winsize.ws_row == rows)
    return 2;  /* No resizing necessary */

  if(get_privileged_ports(0, &dev)) {
    perror("get_privileged_ports");
    return 0;
  }

  /* Open the console device */
  err = device_open(dev, D_WRITE, "console", &console_dev);
  mach_port_deallocate(mach_task_self(), dev);
  if(err) {
    perror("device_open(console)");
    return 0;
  }

  sz.cols=cols;
  sz.lines=rows;
  err = device_set_status(console_dev, KDRESIZE, (int *)&sz, sizeof(sz));
  device_close(console_dev);
  if(err) {
    perror("device_set_status(KDRESIZE)");
    return 0;
  }

  winsize.ws_col=cols;
  winsize.ws_row=rows;
  if(ioctl(STDOUT_FILENO, TIOCSWINSZ, &winsize)) {
    perror("Set Window Size");
    return 0;
  }

  return 1;  /* Successful */
}

#elif defined(DOS)

/*
 * DOS resizing code for SVGATextMode (c) 1995 Stephen Lee
 * April 97: Screen/font-restoring by <A.Borrmann@tu-bs.de>
 */

#include <pc.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/farptr.h>
#include <string.h>

#define BIOS_COLS_PTR    0x44a
#define BIOS_ROWS_PTR    0x484
#define BIOS_FONTSZ_PTR  0x485
#define BIOS_VGASIZE_PTR 0x44c

// Resizes the DOS Screen.
// Returns 0 on failure, 1 if resized, 2 if no resizing was necessary.
//
int resize_tty(int cols, int rows)
{
  unsigned short old_cols = _farpeekw(_dos_ds, BIOS_COLS_PTR);
  unsigned char  old_rows = _farpeekb(_dos_ds, BIOS_ROWS_PTR)+1;
  unsigned short copy[old_cols*old_rows];
  int x, y, x_min, y_min;

  if(ScreenCols() == cols && ScreenRows() == rows)
    return 2;  /* No resizing necessary */

  dosmemget(0xB8000, old_cols*old_rows*2, copy);

  x_min=(old_cols<cols)?old_cols:cols;
  y_min=(old_rows<rows)?old_rows:rows;
  for(y=0;y<y_min;y++) {
    dosmemput(&(copy[old_cols*y]), 2*x_min, 0xB8000+2*cols*y);
    /* Clear remainder of line when enlarging columns */
    for(x=x_min;x<cols;x++)
      _farpokew(_dos_ds, 0xB8000+2*(cols*y+x), 0x0720);
  }

  /* Clear bottom of screen when enlarging rows */
  for(y=y_min;y<rows;y++)
    for(x=0;x<cols;x++)
      _farpokew(_dos_ds, 0xB8000+2*(cols*y+x), 0x0720);

  /* Write new cols/rows to BIOS */
  _farpokew(_dos_ds, BIOS_COLS_PTR, cols);
  _farpokeb(_dos_ds, BIOS_ROWS_PTR, rows-1);

  /* VGA screen size in bytes (Watcom Debugger uses this variable) */
  _farpokew(_dos_ds, BIOS_VGASIZE_PTR, cols*rows*2);

  return 1;  /* Successful */
}

// Loads standard text font from BIOS ROM.
// Fontsize can be either 8, 14, or 16 (the default).
//
void load_bios_font(int fontsize)
{
  _go32_dpmi_registers regs;

  // Load ROM-font, 16 pixel height as default
  memset(&regs, 0, sizeof(regs));
  switch(fontsize) {
    case  8: regs.x.ax=0x1102; break;
    case 14: regs.x.ax=0x1101; break;
    default: regs.x.ax=0x1104; break;
  }
  _go32_dpmi_simulate_int(0x10, &regs);

  _farpokeb(_dos_ds, BIOS_FONTSZ_PTR, fontsize);
}

#endif  /* DOS */
