/* stm.c - SVGATextMode Main Program */

#include "externs.h"
#include "vga/vga.h"

#define CONFIG_DIR "/etc"
#define FONT_PROG  "setfont"

#define VERSION      "1.84"
#define RELEASE_DATE "Dec 12 2003"

static struct modeline *synthesize_mode_timings(char *modeline);
static void print_modeline(struct modeline *);
static int  valid_mode(struct modeline *mode);
static void list_supported_modes(void);
static void set_font(void);

// Default Textmode Parameters
int width=80, height=25, fontwidth=9, fontheight=16, refresh=70000;
int cursor_start=-1, cursor_end=-1, underline=-1, border_color=-1;
int blink=-1, isofont=-1, lineart=-1, palette=-1;
int maxclk=100000, options, cardnum=1, chipset=-1, clockchip=-1;

char *modeline, *fontprog, *resetprog;

int verbose, force;


int main(int argc, char *argv[])
{
  struct modeline *mode=NULL;
  char *cmd=*argv, *configfile=NULL;
  int clock, horiz, vert, invalid;
  int c, init=0, grab=0, list=0, noprog=0;

  while((c = getopt(argc, argv, "#:c:fghilnsvV?")) != -1) {
    switch(c) {
    case '#':  /* Video Card Number */
      cardnum=atoi(optarg);
      break;
    case 'c':  /* Override Config File */
      configfile=optarg;
      break;
    case 'f':  /* Force settings regardless of limits */
      force=1;
      break;
    case 'g':  /* Grab video mode settings */
      grab=1;
      break;
    case 'i':  /* Initialize VGA Card (POST) */
      init=1;
      break;
    case 'l':  /* List supported modes */
      list=1;
      break;
    case 'n':  /* No programming done */
      noprog=1;
      break;
    case 's':  /* Scan PCI Bus Only */
      pci_find_vendor_vga(1);
      exit(0);
    case 'v':  /* Be verbose; show version information */
      verbose++;
      break;
    case ':':
      fprintf(stderr, "%s: option requires an argument -- %c\n", cmd, optopt);
      exit(1);
    case 'h':
    case '?':
      printf("Usage: %s [-# cardnum] [options] display-mode\n\n"
        "  display-mode can be in any form matching: WxHxFWxFH@R\n"
        "  where W=Width, H=Height, FW=Fontwidth, FH=Fontheight, R=Refresh (in Hz)\n"
        "  Default mode: 80x25x8x16@70 (Defaults specified in config file)\n\n"
        "Available options:\n"
        "  -# cardnum     Specify VGA card number (multiple cards only)\n"
        "  -c configfile  Read system configuration from `configfile'\n"
        "  -f             Ignore built-in restrictions and card limits\n"
        "  -g             Grab the modeline from the current display\n"
        "  -h             Display this help screen\n"
        "  -i             Initialize VGA card to power-on state\n"
        "  -l             List valid modes defined in the TextConfig file\n"
        "  -n             Dry-run; don't actually program card\n"
        "  -s             Scan PCI bus only; lists supported video cards\n"
        "  -v             Be verbose; -vv prints debugging messages\n"
        "  -V             Print version information and exit\n\n", cmd);
    case 'V':
      printf("SVGATextMode version %s (%s)\n", VERSION, RELEASE_DATE);
      exit(0);
    }
  }
  argv+=optind;
  argc-=optind;

  /* Read TextConfig configuration file(s) */
  if(configfile) {
    if(!read_config(configfile)) {
      perror(configfile);
      exit(1);
    }
  } else {
    if(!read_config(CONFIG_DIR"/TextConfig"))
#ifndef DOS
    if(!read_config("~/.TextConfig"))
#endif
    read_config("TextConfig");
  }

  if(!fontprog)
    fontprog=FONT_PROG;

  if(verbose)
    vga_verbose=1;

  /* If no allowable Horiz/Vert Refresh ranges are specified, set defaults */
  if(!horiz_ranges) {
    horiz_ranges=malloc(sizeof(struct range));
    horiz_ranges->next=NULL;
    horiz_ranges->lo=29500;
    horiz_ranges->hi=32500;
  }
  if(!vert_ranges) {
    vert_ranges=malloc(sizeof(struct range));
    vert_ranges->next=NULL;
    vert_ranges->lo=49500;
    vert_ranges->hi=80500;
  }

  /* Get modeline from command line */
  if(argc)
    modeline=*argv;

  /* If -i option is given without a mode, resize tty for 80x25 */
  if(!argc && init)
    resize_tty(80, 25);

  if(chipset != -1)
    vga_chipset=chipset;
  if(clockchip != -1)
    vga_clockchip=clockchip;
  vga_options=options;
  if(!vga_init_card(cardnum, init))
    exit(1);

  /* Fix font after initializing card, if not setting a new mode */
  if(!argc && init) {
    set_font();
    exit(0);
  }

  /* Pick up font cell size from VGA card */
  fontwidth=vga_get_text_cell_width();
  fontheight=vga_get_text_cell_height();

  /* List modes supported by VGA card? */
  if(list) {
    list_supported_modes();
    exit(0);
  }

  /* Print current display mode? */
  if(grab) {
    struct modeline mode={};
    char buf[48];

    vga_get_timing(&mode.timing[0], &mode.timing[1], &mode.timing[2],
                   &mode.timing[3], &mode.timing[4], &mode.timing[5],
                   &mode.timing[6], &mode.timing[7]);
    mode.clock=vga_get_clock();

    sprintf(buf, "%dx%dx%d", mode.timing[0]/8, mode.timing[4]/fontheight,
            fontwidth);
    mode.name=strdup(buf);
    mode.width=fontwidth;
    mode.height=fontheight;

    print_modeline(&mode);
    exit(0);
  }

  /* Search for the first valid modeline requested */
  if(modeline) {
    for(invalid=0,mode=modelist;mode;mode=mode->next)
      if(!strcasecmp(mode->name, modeline)) {
        if(force || valid_mode(mode))
          break;
        invalid=1;
      }

    if(!mode && invalid) {
      fprintf(stderr, "Error: Mode '%s' is not supported by your video "
              "card.\n", modeline);
      exit(1);
    }
  }

  if(!modeline || (!mode && !(mode=synthesize_mode_timings(modeline)))) {
    if(modeline)
      fprintf(stderr, "Error: Mode '%s' could not be found in the config "
              "file.\n", modeline);
    else
      fprintf(stderr, "Error: No default display-mode is available.\n");
    exit(1);
  }

  fontwidth=mode->width;
  fontheight=mode->height;

  /* Check for invalid mode timings */
  invalid=0;
  if(mode->timing[0] < 128 ||
     mode->timing[1] < mode->timing[0] ||
     mode->timing[2] < mode->timing[1] ||
     mode->timing[3] < mode->timing[2])
    invalid=1;
  else if(mode->timing[4] < 96 ||
          mode->timing[5] < mode->timing[4] ||
          mode->timing[6] < mode->timing[5] ||
          mode->timing[7] < mode->timing[6])
    invalid=2;

  if(verbose || invalid)
    printf("Mode '%s': PixelCLK=%d.%03dMHz, %4d %4d %4d %4d  %4d %4d %4d %4d\n",
           mode->name, mode->clock/1000, mode->clock%1000,
           mode->timing[0], mode->timing[1], mode->timing[2], mode->timing[3],
           mode->timing[4], mode->timing[5], mode->timing[6], mode->timing[7]);
  if(invalid) {
    fprintf(stderr, "Error: %s mode timings are out of bounds or invalid.\n",
            invalid==1?"Horizontal":"Vertical");
    exit(1);
  }

  /* Calculate Horizontal/Vertical refresh rates */
  clock=mode->clock*1000;
  if(mode->width == 9)
    clock=clock*8/9;

  horiz=(!mode->timing[3])?0 : (clock/mode->timing[3]);
  vert=(!mode->timing[7])?0 : (horiz*1000/mode->timing[7]);
  invalid=(horiz < 11500)?1:(vert < 49500)?2:0;
   
  /* Check for invalid refresh rates */
  if(!force && (verbose || invalid))
    printf("Refresh Rate Selected: %d.%02dkHz/%d.%02dHz\n",
           horiz/1000, horiz%1000/10, vert/1000, vert%1000/10);
  if(!force && invalid) {
    fprintf(stderr, "Error: %s Refresh rate too low!\n",
            invalid==1?"Horizontal":"Vertical");
    exit(1);
  }

  /* If this is a dry-run, then stop here */
  if(noprog)
    exit(0);

  /* Don't interrupt the VGA card in the middle of the screen */
  vga_wait_for_retrace();

  /* Set new clock frequency; bail out if unable to set */
  if(!vga_set_clock(mode->clock)) {
    fprintf(stderr, "Error: Unable to set VGA Pixel Clock to %d.%03d MHz!\n",
            mode->clock/1000, mode->clock%1000);
    exit(1);
  }

  /* Round off vertical display area to match font height */
  mode->timing[4] -= mode->timing[4] % (fontheight*
                      ((mode->flags & DOUBLESCAN)?2:1));

  /* Resize TTY Rows x Cols */
  if(!resize_tty(mode->timing[0]/8, mode->timing[4]/fontheight/
                 ((mode->flags & DOUBLESCAN)?2:1)))
    exit(0);

  /* Set new VGA mode timings */
  vga_set_timing(mode->timing[0], mode->timing[1], mode->timing[2],
                 mode->timing[3], mode->timing[4], mode->timing[5],
                 mode->timing[6], mode->timing[7]);
  printf("Set Mode '%s': PixelCLK=%d.%03dMHz, %4d %4d %4d %4d  %4d %4d %4d %4d\n",
         mode->name, mode->clock/1000, mode->clock%1000,
         mode->timing[0], mode->timing[1], mode->timing[2], mode->timing[3],
         mode->timing[4], mode->timing[5], mode->timing[6], mode->timing[7]);

  /* Set sync polarity, doublescan, interlace... */
  vga_set_flags(~mode->flags, 0);
  vga_set_flags(mode->flags, 1);

  /* Read back timings to make sure they're set properly */
  vga_get_timing(&mode->timing[0], &mode->timing[1], &mode->timing[2],
                 &mode->timing[3], &mode->timing[4], &mode->timing[5],
                 &mode->timing[6], &mode->timing[7]);
  printf("Get Mode '%s': PixelCLK=%d.%03dMHz, %4d %4d %4d %4d  %4d %4d %4d %4d\n",
         mode->name, mode->clock/1000, mode->clock%1000,
         mode->timing[0], mode->timing[1], mode->timing[2], mode->timing[3],
         mode->timing[4], mode->timing[5], mode->timing[6], mode->timing[7]);

  /* Set new cell size */
  vga_set_text_cell_width(fontwidth);
  vga_set_text_cell_height(fontheight);

  if(options & LOADFONT)
    set_font();

  /* Set extended text registers */
  if(underline != -1)
    vga_set_text_underline(underline);
  if(border_color != -1)
    vga_set_dac_entry(16, border_color);
  if(cursor_start != -1)
    vga_set_text_cursor_size(cursor_start*fontheight/32,
                             cursor_end*fontheight/32);

  if(blink != -1)   vga_set_text_attr(VGA_TEXT_BLINK, blink);
  if(isofont != -1) vga_set_text_attr(VGA_TEXT_FONT_512, isofont);
  if(lineart != -1) vga_set_text_attr(VGA_TEXT_DUP_PIXEL9, lineart);
  if(palette != -1) vga_set_text_attr(VGA_TEXT_8BIT_DAC, palette);

  /* Run user reset script */
  if(resetprog && *resetprog) {
    char buf[8192];

    strmaxcpy(buf, resetprog, 8100);
    sprintf(buf+strlen(buf), " %d %d", fontwidth, fontheight);
    system(resetprog);
  }

  exit(0);
}

static struct modeline *synthesize_mode_timings(char *modeline)
{
  static struct modeline mode;
  char *s=modeline;

  if(*s) {
    if(!isdigit(*s))
      return NULL;

    /* Width */
    width=atoi(s);
    for(;isdigit(*s);s++);

    if((*s|32) == 'x') {
      if(!isdigit(*++s))
        return NULL;

      /* Height */
      height=atoi(s);
      for(;isdigit(*s);s++);
    }

    if((*s|32) == 'x') {
      if(!isdigit(*++s))
        return NULL;

      /* Font Width */
      fontwidth=atoi(s);
      for(;isdigit(*s);s++);
    }

    if((*s|32) == 'x') {
      if(!isdigit(*++s))
        return NULL;

      /* Font Height */
      fontheight=atoi(s);
      for(;isdigit(*s);s++);
    }

    if(*s == '@') {
      if(!isdigit(*++s))
        return NULL;

      /* Vertical Refresh Rate (Hz) */
      refresh=atof(s)*1000;
      for(;isdigit(*s) || *s == '.';s++);
    }

    if(*s)
      return NULL;
  }

  if(width < 40 || width > 255) {
    fprintf(stderr, "Error: Textmode width can only be between 40 and 255 "
            "characters.\n");
    exit(1);
  } if(width % 2) {
    fprintf(stderr, "Error: SVGATextMode only supports text widths that are a "
            "multiple of 2.\n");
    exit(1);
  } if(height < 10 || height > 255) {
    fprintf(stderr, "Error: Textmode height can only be between 10 and 255 "
            "characters.\n");
    exit(1);
  } if((fontwidth != 8 && fontwidth != 9) ||
       fontheight < 1 || fontheight > 32) {
    fprintf(stderr, "Error: Invalid character font size %dx%d.\n", fontwidth,
            fontheight);
    exit(1);
  } if(!force && (refresh < 50000 || refresh > 250000)) {
    fprintf(stderr, "Error: Vertical Refresh rate must be between "
            "50 and 250 Hz.\n");
    exit(1);
  }

  mode.name=modeline;
  mode.width=fontwidth;
  mode.height=fontheight;

  // Reference Mode '80x25x9': 28.300MHz,  640 680 776 800  400 412 414 449

  /* Calculate Horizontal Timings */
  mode.timing[0]=width*8;
  mode.timing[1]=mode.timing[0]+(((mode.timing[0]-40)/80) & ~7)+40;
  mode.timing[2]=mode.timing[0]+((mode.timing[0]*3/16) & ~7)+16;
  mode.timing[3]=mode.timing[0]+((mode.timing[0]*31/128+5) & ~7);

  /* Calculate Vertical Timings */
  mode.timing[4]=height*fontheight;
  mode.timing[5]=mode.timing[4]+12;
  mode.timing[6]=mode.timing[4]+14;
  mode.timing[7]=mode.timing[4]+49;

  /* Adjust Horizontal based on kHz needed */
  mode.timing[1]+=((mode.timing[7]-450)/100*8) & ~7;
  mode.timing[3]+=((mode.timing[0]+160)/33*(mode.timing[7]-409)/304) & ~7;
  mode.timing[1]+=((refresh-70000)/4000) & ~7;
  mode.timing[3]+=(refresh/1000-70) & ~7;

  /* Calculate pixel clock necessary for given Vertical Refresh Rate */
  mode.clock=mode.timing[3]*refresh/1000*mode.timing[7]/1000+5;
  if(mode.width == 9) {
    mode.clock=mode.clock*9/8;
    if(mode.clock > maxclk*9/8)
      mode.clock=maxclk*9/8;
  } else if(mode.clock > maxclk)
    mode.clock=maxclk;

  if(!force && !valid_mode(&mode)) {
    fprintf(stderr, "Error: Mode '%s' is not supported by your video "
            "card.\n", modeline);
    exit(1);
  }

  return &mode;
}

static void print_modeline(struct modeline *mode)
{
  char buf[48], buf2[32], buf3[256], *s=buf3;
  int i, clock, horiz, vert;

  clock=mode->clock*1000;
  if(mode->width == 9)
    clock=clock*8/9;
  horiz=(!mode->timing[3])?0 : (clock/mode->timing[3]);
  vert=(!mode->timing[7])?0 : (horiz*1000/mode->timing[7]);

  sprintf(buf, "\"%.45s\"", mode->name);
  sprintf(buf2, "%d.%03d", mode->clock/1000, mode->clock%1000);
  *buf3='\0';
  for(i=0;modeline_flags[i];i++)
    if(mode->flags & (1 << i))
      s+=sprintf(s, "%s%s", s==buf3?"":" ", modeline_flags[i]);
  printf("%-13s %7s  %4d %4d %4d %4d   %4d %4d %4d %4d  font %dx%-2d %s  # "
         "%d.%03dkHz/%d.%02dHz\n", buf, buf2,
         mode->timing[0], mode->timing[1], mode->timing[2], mode->timing[3],
         mode->timing[4], mode->timing[5], mode->timing[6], mode->timing[7],
         mode->width, mode->height, buf3, horiz/1000, horiz%1000,
         vert/1000, vert%1000);
}

static int valid_mode(struct modeline *mode)
{
  struct range *range;
  int clock, horiz, vert;

  clock=mode->clock*1000;
  if(mode->width == 9)
    clock=clock*8/9;
  if(clock/1000 > maxclk+250)
    return 0;

  horiz=(!mode->timing[3])?0 : (clock/mode->timing[3]);
  vert=(!mode->timing[7])?0 : (horiz*1000/mode->timing[7]);

  /* Check horizontal ranges */
  for(range=horiz_ranges;range;range=range->next)
    if(horiz >= range->lo && horiz <= range->hi)
      break;
  if(!range)
    return 0;

  /* Check vertical ranges */
  for(range=vert_ranges;range;range=range->next)
    if(vert >= range->lo && vert <= range->hi)
      break;
  return range?1:0;
}

static void list_supported_modes()
{
  struct modeline *mode;
  int count=0;

  for(mode=modelist;mode;mode=mode->next)
    if(force || valid_mode(mode)) {
      print_modeline(mode);
      count++;
    }
  
  if(!count)
    printf("No valid modes are defined for VGA Card #%d.\n", cardnum);
}

/* Traverse the font list for the first font that matches the cell size */
static void set_font()
{
  struct font *ptr;
  char buf[8192];
  int len;

  if(!fontprog || !*fontprog)
    return;

  for(ptr=fontlist;ptr;ptr=ptr->next)
    if(ptr->width == fontwidth && ptr->height == fontheight)
      break;
  if(!ptr)
    ptr=fontlist;

  strmaxcpy(buf, fontprog, 8190);
  len=strlen(buf);
  if(ptr) {
    buf[len++]=' ';
    strmaxcpy(buf+len, ptr->name, 8192-len);
  }
  system(buf);
}
