/* ati.c */

#include "vga.h"

struct pcimodel model_ati[]={
  {0x1002, 0x4c57, 0x4c5a, 4,  "Radeon Mobility"},
  {0x1002, 0x5040, 0x505f, 3,  "Rage 128 P* Series"},
  {0x1002, 0x5140, 0x515f, 4,  "Radeon Q* Series"},
  {0x1002, 0x5160, 0x516f, 4,  "Radeon R200"},
  {0x1002, 0x5240, 0x525f, 3,  "Rage 128 R* Series"},
  {0x1002, 0x5340, 0x535f, 3,  "Rage 128 S* Series"},
  {0x1002, 0x5440, 0x545f, 3,  "Rage 128 Pro T* Series"},
  {0}
};

#define MAX_PCLK 350000  // Max Pixel Clock, kHz
#define MIN_PCLK 7500    // Min Pixel Clock, kHz

static void (*ati_set_timing_fallback)
  (int h_display, int h_start, int h_end, int h_total,
   int v_display, int v_start, int v_end, int v_total);
static void (*ati_set_text_attr_fallback)(int flags, int enable);
static int  (*ati_get_text_attr_fallback)(void);


#define MMIO(index) (*(int *)&mmio[index])

unsigned char *mmio;

static void change_ati(void);


static int ati_inpll(int addr)
{
  while((MMIO(0x0008) & 0xbf) != addr)
    MMIO(0x0008)=addr & 0x3f;
  return MMIO(0x000c);
}

static void ati_outpll(int addr, int val)
{
  while((MMIO(0x0008) & 0xbf) != (addr | 0x80))
    MMIO(0x0008)=(addr & 0x3f) | 0x80;
  MMIO(0x000c)=val;
}

/* Initialize Chipset */
int ati_init(int card, int index)
{
  unsigned int address=card_list[card].address;
  int fd, memory;

  iopl(3);

  /* Set up memory maps */
  if((fd=open("/dev/mem", O_RDWR)) == -1) {
    perror("/dev/mem");
    return -1;
  }

  /* Determine chipset type */
  if(vga_verbose)
    printf("Detected ATI chipset (%s)\n", model_ati[index].name);

  /* Map I/O Ports */
  mmio=mmap(0, 0x2000, PROT_READ|PROT_WRITE, MAP_SHARED, fd, address);
  if(!mmio) {
    perror("mmap failed");
    close(fd);
    return -1;
  }

  change_ati();

  /* Detect onboard memory size (in KB) */
  memory=MMIO(0x00f8) / 1024;
  if(!memory)
    memory=8192;  /* Some boards report 0 if they have 8 MB */

  if(vga_verbose)
    printf("Avail memory: %d KB\n", memory);

  /* Set globals */
  card_list[card].chipset=VGA_CS_ATI;
  card_list[card].memory=memory;
  card_list[card].maxclock=MAX_PCLK;

  vga_chipset=VGA_CS_ATI;
  vga_memory=memory;
  vga_maxclock=MAX_PCLK;

  /* Unlock the CRTC registers */
  vga_outcrtc(0x11, vga_incrtc(0x11) & 0x7f);

  if(vga_verbose > 1)
    printf("Current PLL Timings: rd=%d, fd=%d, pd=%d\n",
           ati_inpll(0x03) & 0x03ff, ati_inpll(0x05) & 0x07ff,
           (ati_inpll(0x09) & 0x00070000) >> 16);

  return fd;
}

static void ati_blank_screen(int blank)
{
  MMIO(0x0054) = (MMIO(0x0054) & ~0x700) | (blank ? 0x700:0);
}

static void ati_set_palette_rgb(int index, int rgb)
{
  MMIO(0x00b0)=index;
  MMIO(0x00b4)=rgb & 0xffffff;
}

static void ati_set_palette(int index, int r, int g, int b)
{
  MMIO(0x00b0)=index;
  MMIO(0x00b4)=((r & 0xff) << 16)|((g & 0xff) << 8)|(b & 0xff);
}

/* Display start address */

static void ati_set_display_start(int addr)
{
  MMIO(0x0224)=addr;
}

static int ati_get_display_start()
{
  return MMIO(0x0224);
}

/* Logical scanline width (usually multiple of 8) in bytes per pixel */

static void ati_set_display_width(int width)
{
  MMIO(0x022c)=width >> 3;
}

static int ati_get_display_width()
{
  return MMIO(0x022c) << 3;
}

static void ati_set_text_attr(int flags, int enable)
{
  if(flags & VGA_TEXT_8BIT_DAC) {
    MMIO(0x0058)=(MMIO(0x0058) & ~0x100)|(enable ? 0x100:0);
    flags &= ~VGA_TEXT_8BIT_DAC;
  }

  ati_set_text_attr_fallback(flags, enable);
}

static int ati_get_text_attr()
{
  int flags=ati_get_text_attr_fallback() & ~VGA_TEXT_8BIT_DAC;

  if(MMIO(0x0058) & 0x100)
    flags |= VGA_TEXT_8BIT_DAC;
  return flags;
}


static void ati_set_timing(int h_display, int h_start, int h_end, int h_total,
                           int v_display, int v_start, int v_end, int v_total)
{
  /* Are we in textmode? */
  if(1) {
    /* This driver does not support textmode calculations!
       Must use standard vga registers instead. */
    ati_set_timing_fallback(h_display, h_start, h_end, h_total,
                            v_display, v_start, v_end, v_total);
    return;
  }

  h_display/=8;

  /* Set default Logical Width based on Bits Per Pixel mode selected */
  /* FIX: Currently Unknown */
  vga_set_display_width(h_display*4);

  /* Horizontal Total & Display */
  h_display--;
  h_total=h_total/8-5;
  MMIO(0x0200)=(MMIO(0x0200) & ~0x1ff03ff) |
               (h_total & 0x3ff)|((h_display & 0x1ff) << 16);

  /* Horizontal Sync Start & End (Comparison) */
  h_start=(h_start-8) & ~7;
  h_end=h_end/8-1;
  MMIO(0x0204)=(MMIO(0x0204) & ~0x3f1ff8) |
               (h_start & 0x1ff8) | ((h_end & 0x3f) << 16);

  /* Vertical Total & Display */
  v_total-=2;
  v_display-=1;
  MMIO(0x0208)=(MMIO(0x0208) & ~0x7ff07ff) |
               (v_total & 0x7ff)|((v_display & 0x7ff) << 16);

  /* Vertical Sync Start & End (Comparison) */
  MMIO(0x020c)=(MMIO(0x020c) & ~0x1f07ff) |
               (v_start & 0x7ff) | ((v_end & 0x1f) << 16);
}

static void ati_get_timing(int *h_display,int *h_start,int *h_end,int *h_total,
                           int *v_display,int *v_start,int *v_end,int *v_total)
{
  int h_sync, v_sync;

  *h_total=((MMIO(0x0200) & 0x3ff)+5) * 8;
  *h_display=(((MMIO(0x0200) >> 16) & 0x1ff)+1) * 8;
  h_sync=(MMIO(0x0204) & 0x1fff) >> 3;
  *h_start=(h_sync+1) * 8;
  *h_end=((MMIO(0x0204) >> 16) & 0x3f);
  *h_end=((h_sync & ~0x3f)|(*h_end & 0x3f))+
          (((h_sync & 0x3f) > *h_end)?0x40:0x00);
  *h_end=(*h_end+1) * 8;
  
  *v_total=(MMIO(0x0208) & 0x7ff)+2;
  *v_display=((MMIO(0x0208) >> 16) & 0x7ff)+1;
  v_sync=(MMIO(0x020c) & 0x7ff);
  *v_start=v_sync;
  *v_end=((MMIO(0x020c) >> 16) & 0x1f);
  *v_end=((v_sync & ~0x1f)|(*v_end & 0x1f))+
          (((v_sync & 0x1f) > *v_end)?0x20:0x00);
}

static void ati_write_pll_regs(int ppll_ref_div, int ppll_div_3,
                               int htotal_cntl)
{
  int reg, graphics=0;

  /* reset vclk ecp controller */
  ati_outpll(0x08, ati_inpll(0x08) & ~0x03);

  /* pll div select */
  while((MMIO(0x0008) & 0x300) != 0x300)
    MMIO(0x0008) |= 0x300;

  /* reset pll controller */
  ati_outpll(0x02, ati_inpll(0x02) | 0x0001);

  while(((reg=ati_inpll(0x03)) & 0x03ff) != (ppll_ref_div & 0x03ff))
    ati_outpll(0x03, (reg & ~0x03ff) | ppll_ref_div);

  /* 0x05 = VGA Text Clock, 0x07 = CRTC1 Graphics Clock */
  if(graphics) {
    while(((reg=ati_inpll(0x07)) & 0x000707ff) != (ppll_div_3 & 0x000707ff))
      ati_outpll(0x07, (reg & ~0x000707ff) | ppll_div_3);
  } else {
    while(((reg=ati_inpll(0x05)) & 0x000707ff) != (ppll_div_3 & 0x000707ff))
      ati_outpll(0x05, (reg & ~0x000707ff) | ppll_div_3);
  }

  ati_outpll(0x09, htotal_cntl);

  /* clear reset pll controller */
  ati_outpll(0x02, ati_inpll(0x02) & ~0x0001);

  /* clear reset vclk ecp controller */
  ati_outpll(0x08, ati_inpll(0x08) | 0x03);

  if(vga_verbose > 1)
    printf("Setting PLL Timings: rd=%d, fd=%d, pd=%d\n",
           ati_inpll(0x03) & 0x03ff, ati_inpll(graphics?0x07:0x05) & 0x07ff,
           (ati_inpll(0x09) & 0x00070000) >> 16);
}

// From RAGE 128 VR/RAGE 128 GL Register Reference Manual (Technical
// Reference Manual P/N RRG-G04100-C Rev. 0.04), page 3-17 (PLL_DIV_[3:0]).
// Values := VCLK_SRC/x where x = post_divs[0..7].
//
static int post_divs[8]={1, 2, 4, 8, 3, 16, 6, 12};

static int ati_set_clock(int freq)
{
  int i, pll_output_freq=0, feedback_div;

  freq /= 10;
  if(freq > 35000)     /* Max PLL frequency: 35000 */
    return 0;
  if(freq < 12000/16)  /* Min PLL frequency: 12000 */
    return 0;

  for(i=0;i<8;i++) {
    pll_output_freq = post_divs[i] * freq;
    if(pll_output_freq >= 12000 && pll_output_freq <= 35000)
      break;
  }

  /* reference freq: 2700 */
  feedback_div = (60 * pll_output_freq + (2700/2)) / 2700;

  ati_write_pll_regs(60, feedback_div | (i << 16), 0);
  return freq;
}

static int ati_get_clock()
{
  // To be written eventually. Patches welcome!
  return 0;
}

/* VGA Flags */

static void ati_set_flags(int flags, int enable)
{
  if(flags & (VGA_DOUBLESCAN|VGA_INTERLACE|VGA_CSYNC)) {
    int reg=MMIO(0x0050);

    if(flags & VGA_DOUBLESCAN)
      reg=(reg & ~0x01) | (enable ? 0x01:0);
    if(flags & VGA_INTERLACE)
      reg=(reg & ~0x02) | (enable ? 0x02:0);
    if(flags & VGA_CSYNC)
      reg=(reg & ~0x10) | (enable ? 0x10:0);

    MMIO(0x0050)=reg;
  }

  if(flags & VGA_NHSYNC)
    MMIO(0x0204)=(MMIO(0x0204) & ~0x800000) | (enable ? 0x800000:0);
  if(flags & VGA_NVSYNC)
    MMIO(0x020c)=(MMIO(0x020c) & ~0x800000) | (enable ? 0x800000:0);
}

static int ati_get_flags()
{
  int flags=0, reg=MMIO(0x0050);

  if(reg & 0x01)
    flags |= VGA_DOUBLESCAN;
  if(reg & 0x02)
    flags |= VGA_INTERLACE;
  if(reg & 0x10)
    flags |= VGA_CSYNC;
  if(MMIO(0x0204) & 0x800000)
    flags |= VGA_NHSYNC;
  if(MMIO(0x020c) & 0x800000)
    flags |= VGA_NVSYNC;

  return flags;
}

static void ati_wait_for_retrace()
{
  int i;

  MMIO(0x0044) = 0x04;
  for(i=0;i < VGA_TIMEOUT && (MMIO(0x0044) & 0x04);i++);
}


static void change_ati()
{
  change_std();

  vga_set_palette = ati_set_palette;
  vga_set_palette_rgb = ati_set_palette_rgb;

  vga_set_display_start = ati_set_display_start;
  vga_set_display_width = ati_set_display_width;

  vga_get_display_start = ati_get_display_start;
  vga_get_display_width = ati_get_display_width;

  ati_set_text_attr_fallback = vga_set_text_attr;
  ati_get_text_attr_fallback = vga_get_text_attr;
  vga_set_text_attr = ati_set_text_attr;
  vga_get_text_attr = ati_get_text_attr;

  ati_set_timing_fallback = vga_set_timing;
  vga_set_timing = ati_set_timing;
  vga_get_timing = ati_get_timing;

  vga_set_clock = ati_set_clock;
  vga_get_clock = ati_get_clock;

  vga_set_flags = ati_set_flags;
  vga_get_flags = ati_get_flags;

  vga_blank_screen = ati_blank_screen;
  vga_wait_for_retrace = ati_wait_for_retrace;
}
