/* pci.c - Routines for talking directly to the PCI bus */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/io.h>

#include "pcispec.h"
#include "pci.h"

struct pcicard card_list[NUM_CARDS];


static int pci_scan_bus(int class)
{
  int offset, value, num=0, vendor, device;

  for(device=0;device < 0x1000 && num < NUM_CARDS;device += 0x8) {
    vendor=pci_read_config_dword(device, 0);
    if(vendor == -1)
      continue;
    if(pci_read_config_word(device, 10) != class)
      continue;

    card_list[num].device=device;
    card_list[num].vendor=vendor;

    /* Find I/O Ports, Framebuffer, and Control Addresses */
    for(offset=16;offset < 40;offset+=4) {
      if(!(value=pci_read_config_dword(device, offset)))
        continue;
      if(value & 0x1) {
        if(!card_list[num].ioaddr)        /* I/O Ports */
          card_list[num].ioaddr = value & ~0x3;
      } else if(value & 0x8) {
        if(!card_list[num].memory)        /* Framebuffer */
          card_list[num].memory = value & ~0xf;
      } else if(!card_list[num].address)  /* Control Address */
        card_list[num].address = value & ~0xf;
    }

    if(card_list[num].address && !card_list[num].memory) {
      card_list[num].memory=card_list[num].address;
      card_list[num].address=0;
    }

    num++;  /* Add card to list */
  }

  return num;
}

int pci_find_vendor_vga(int verbose)
{
  int i, num;

  if(iopl(3) == -1) {
    perror("PCI access");
    return 0;
  }

  if(verbose)
    printf("Scanning PCI Bus...\n");

  num=pci_scan_bus(PCI_CLASS_DISPLAY_VGA);
  iopl(0);

  if(!num) {
    printf("No video cards were detected in your system.\n");
    return 0;
  }

  if(verbose)
    for(i=0;i < num;i++)
      printf("  %02x:%02x.%x  VGA Card #%d Found: Address=%08x, Memory=%08x\n",
             card_list[i].device >> 8, PCI_SLOT(card_list[i].device),
             PCI_FUNC(card_list[i].device), i+1, card_list[i].address,
             card_list[i].memory);

  return num;
}

static int pci_find_capability(int dev, int cap)
{
  unsigned char pos, id;
  int ttl=48;

  /* Check that the card supports capability lists */
  if(!(pci_read_config_word(dev, PCI_STATUS) & PCI_STATUS_CAP_LIST))
    return 0;

  pos=pci_read_config_byte(dev, PCI_CAPABILITY_LIST);
  while(ttl-- && pos >= 0x40) {
    pos &= ~3;
    id=pci_read_config_byte(dev, pos + PCI_CAP_LIST_ID);
    if(id == 0xff)
      break;
    if(id == cap)
      return pos;
    pos=pci_read_config_byte(dev, pos + PCI_CAP_LIST_NEXT);
  }

  return 0;
}

/* Initialize a PCI device for I/O. Returns 1 if the card was already
   initialized, 0 otherwise */
int pci_enable_device(int device)
{
  int pm, pwr_command;
  short cmd;

  if(iopl(3) == -1) {
    perror("PCI access");
    exit(1);
  }

  /* Set power management state of PCI to D0 */
  if((pm=pci_find_capability(device, PCI_CAP_ID_PM))) {
    pwr_command=pci_read_config_word(device, pm+PCI_PM_CTRL);
    pci_write_config_word(device, pm+PCI_PM_CTRL,
           (pwr_command & ~PCI_PM_CTRL_STATE_MASK));
  }

  cmd=pci_read_config_word(device, PCI_COMMAND);
  if(!(cmd & PCI_COMMAND_MEMORY) || !(cmd & PCI_COMMAND_IO)) {
    pci_write_config_word(device, PCI_COMMAND,
        cmd | PCI_COMMAND_MEMORY | PCI_COMMAND_IO);
    cmd=0;
  } else
    cmd=1;

  iopl(0);
  return cmd;
}
