/* stdvga.c */

#include "vga.h"

struct pcimodel model_svga[]={
  {0x1002, 0x4756, 0,      VGA_CS_ATIMACH64, "ATI 3D Rage 215IIC"},
  {0x100c, 0x3202, 0x3207, VGA_CS_ET4000,    "Tseng ET4000", /*VGA_CLOCK_ET4000*/},
  {0x100c, 0x3208, 0,      VGA_CS_ET6000,    "Tseng ET6000", VGA_CLOCK_ET6000},
  {0x5333, 0x8811, 0,      VGA_CS_S3,        "S3 Trio64", VGA_CLOCK_S3TRIO},
  {0x5333, 0x8a01, 0,      VGA_CS_S3,        "S3 ViRGE/GX", VGA_CLOCK_S3VIRGE},
  {0}
};

unsigned int crtc_index;

#define OUT_CRTC(index, data)  (outw(((data) << 8) | (index), crtc_index+4))
#define IN_CRTC(index)         (outb(index, crtc_index+4), inb(crtc_index+5))

#define OUT_SEQ(index, data)   (outw(((data) << 8) | (index), 0x3c4))
#define IN_SEQ(index)          (outb(index, 0x3c4), inb(0x3c5))

#define OUT_GRA(index, data)  (outw(((data) << 8) | (index), 0x3ce))
#define IN_GRA(index)         (outb(index, 0x3ce), inb(0x3cf))

static int std_inmisc(void);


/* Card Initialization */
int std_init(int card, int index)
{
  change_std();

  if(index != -1) {
    vga_chipset=model_svga[index].rev;
    if(model_svga[index].clockchip)
      vga_clockchip=model_svga[index].clockchip;
    if(vga_verbose)
      printf("Detected SVGA chipset (%s)\n", model_svga[index].name);
  }

  /* Unprotect CRTC registers */
  OUT_CRTC(0x11, IN_CRTC(0x11) & ~0x80);

  if(vga_chipset == VGA_CS_S3) {
    OUT_CRTC(0x39, 0xa5);  // Unprotect system extension regs (CRTC 0x50-0x5E)
    OUT_CRTC(0x38, 0x48);  // S3 register set (CRTC 0x30-0x3C)
    OUT_CRTC(0x35, IN_CRTC(0x35) & ~0x30);  // Vertical/Horizontal timing regs
    OUT_CRTC(0x34, IN_CRTC(0x34) & ~0xa0);  // 8/9 dot clock & MISC Clock bits
    OUT_CRTC(0x40, IN_CRTC(0x40) | 0x01);   // Accelerator command access
  }

  return 1;
} 


static int std_inmisc()
{
  return inb(0x3cc);
}

static void std_outmisc(int val)
{
  outb(val, 0x3c2);
}

static int std_incrtc(int i)
{
  return IN_CRTC(i);
}

static void std_outcrtc(int i, int val)
{
  OUT_CRTC(i, val);
}

static int std_inseq(int i)
{
  return IN_SEQ(i);
}

static void std_outseq(int i, int val)
{
  OUT_SEQ(i, val);
}

static int std_ingra(int i)
{
  return IN_GRA(i);
}

static void std_outgra(int i, int val)
{
  OUT_GRA(i, val);
}

static int std_inattr(int i)
{
  inb(crtc_index+0x0a);  /* reset flip-flop */
  outb(i, 0x3c0);
  return inb(0x3c1);
}

static void std_outattr(int i, int val)
{
  inb(crtc_index+0x0a);  /* reset flip-flop */
  outb(i|0x20, 0x3c0);
  outb(val, 0x3c0);
}

static void std_blank_screen(int blank)
{
  OUT_SEQ(0x01, (IN_SEQ(0x01) & ~0x20) | (blank ? 0x20:0));
}

static void std_set_palette(int index, int r, int g, int b)
{
  outb(index, 0x3c8);
  outb(r, 0x3c9);
  outb(g, 0x3c9);
  outb(b, 0x3c9);
}

static void std_set_palette_rgb(int index, int rgb)
{
  std_set_palette(index, rgb >> 16, rgb >> 8, rgb);
}

static void std_set_palette_vector(int index, int num,
                                   const unsigned char *vector)
{
  int i;

  outb(index, 0x3c8);

  for(i=0;i<num*3;i++)
    outb(vector[i], 0x3c9);
}

static int std_get_palette_rgb(int index)
{
  outb(index, 0x3c8);
  return (inb(0x3c9) << 16) | (inb(0x3c9) << 8) | inb(0x3c9);
}

static void std_get_palette_vector(int index, int num, unsigned char *vector)
{
  int i;

  outb(index, 0x3c8);

  for(i=0;i<num*3;i++)
    vector[i]=inb(0x3c9);
}

static void std_set_dac_entry(int reg, int index)
{
  if(reg >= 0 && reg <= 16)
    std_outattr(reg == 16?0x11:reg, index);
}

static int std_get_dac_entry(int reg)
{
  if(reg >= 0 && reg <= 16)
    return std_inattr(reg == 16?0x11:reg);
  return 0;
}

/* Display start address */

static void std_set_display_start(int addr)
{
  addr >>= 2;

  OUT_CRTC(0x0d, addr);
  OUT_CRTC(0x0c, addr >> 8);
}

static int std_get_display_start()
{
  return (IN_CRTC(0x0d) | (IN_CRTC(0x0c) << 8)) << 2;
}

/* Logical scanline width in bytes (multiple of 8) */

static void std_set_display_width(int width)
{
  OUT_CRTC(0x13, width >> 3);
}

static int std_get_display_width()
{
  return IN_CRTC(0x13) << 3;
}

/* Textmode Features */

static void std_set_text_cursor_size(int start, int end)
{
  OUT_CRTC(0x0a, (IN_CRTC(0x0a) & ~0x1f) | (start & 0x1f));
  OUT_CRTC(0x0b, (IN_CRTC(0x0b) & ~0x1f) | (end & 0x1f));
}

static void std_get_text_cursor_size(int *start, int *end)
{
  *start=IN_CRTC(0x0a) & 0x1f;
  *end=IN_CRTC(0x0b) & 0x1f;
}

static void std_set_text_cursor_location(int address)
{
  OUT_CRTC(0x0f, address);
  OUT_CRTC(0x0e, address >> 8);
}

static int std_get_text_cursor_location()
{
  return (IN_CRTC(0x0e) << 8) | IN_CRTC(0x0f);
}

static void std_set_text_cell_height(int height)
{
  OUT_CRTC(0x09, (IN_CRTC(0x09) & ~0x1f) | ((height-1) & 0x1f));
}

static int std_get_text_cell_height()
{
  return (IN_CRTC(0x09) & 0x1f)+1;
}

static void std_set_text_cell_width(int width)
{
  OUT_SEQ(0x01, (IN_SEQ(0x01) & ~0x01) | (width & 1 ? 0:0x01));
  std_outattr(0x13, (std_inattr(0x13) & ~0x0f) | (width & 1 ? 8:0));
}

static int std_get_text_cell_width()
{
  return (IN_SEQ(0x01) & 0x01)?8:9;
}

static void std_set_text_underline(int scanline)
{
  OUT_CRTC(0x14, (IN_CRTC(0x14) & ~0x1f) | (scanline & 0x1f));
}

static int std_get_text_underline()
{
  return (IN_CRTC(0x14) & 0x1f);
}

static void std_set_text_attr(int flags, int enable)
{
  if(flags & (VGA_TEXT_FONT_512|VGA_TEXT_DUP_PIXEL9|VGA_TEXT_BLINK)) {
    int reg=std_inattr(0x10);

    if(flags & VGA_TEXT_FONT_512)
      reg=(reg & ~0x2) | (enable ? 0x2:0);
    if(flags & VGA_TEXT_DUP_PIXEL9)
      reg=(reg & ~0x4) | (enable ? 0x4:0);
    if(flags & VGA_TEXT_BLINK)
      reg=(reg & ~0x8) | (enable ? 0x8:0);

    std_outattr(0x10, reg);
  }
}

static int std_get_text_attr()
{
  int flags=0, reg=std_inattr(0x10);

  if(reg & 0x2)
    flags |= VGA_TEXT_FONT_512;
  if(reg & 0x4)
    flags |= VGA_TEXT_DUP_PIXEL9;
  if(reg & 0x8)
    flags |= VGA_TEXT_BLINK;

  return flags;
}

/* Set active screen mode size */

static void std_set_timing(int h_display, int h_start, int h_end, int h_total,
                           int v_display, int v_start, int v_end, int v_total)
{
  int h_blanking_start, h_blanking_end;
  int v_blanking_start, v_blanking_end;

  h_display=h_display/8;  /* Display End */
  h_start  =h_start  /8;  /* Sync Start */
  h_end    =h_end    /8;  /* Sync End */
  h_total  =h_total  /8;  /* Total Width */

  if(IN_GRA(0x06) & 0x01) {
    /* Graphics Mode: No border */
    h_blanking_start=h_display;
    h_blanking_end=h_total;
    v_blanking_start=v_start;
    v_blanking_end=v_end;

    /* Set default Logical Width based on Bits Per Pixel mode selected */
    vga_set_display_width(h_display*4);
  } else {
    /* Text Mode: Use overscan border color */
    h_blanking_start=MIN(h_start, h_display+1);
    h_blanking_end=MAX(h_end, h_total-1);
    v_blanking_start=MIN(v_start, v_display+8);
    v_blanking_end=MAX(v_end, v_total-8);

    if(h_blanking_start - h_blanking_end > 63)
      h_blanking_end=h_blanking_start+63;
    if(v_blanking_start - v_blanking_end > 127)
      v_blanking_end=v_blanking_start+127;

    /* Set default Logical Width */
    vga_set_display_width(h_display*4);
  }

  /* If v_total > 1023, turn on double vertical timing (gives us 0 to 2046) */
  if(v_total > 1023) {
    v_display/=2;
    v_start/=2;
    v_end=(v_end+1)/2;     /* Round up so we don't have a zero-height sync */
    v_total=(v_total+1)/2;
    
    /* Set 'Multiply Vertical Registers By Two' */
    OUT_CRTC(0x17, IN_CRTC(0x17) | 0x04);
  } else
    OUT_CRTC(0x17, IN_CRTC(0x17) & ~0x04);  /* Clear Multiply By Two */


  /* Horizontal Total [0..255] (2044) */
  h_total-=5;
  OUT_CRTC(0x00, h_total);

  /* Horizontal Display End [0..255] (2044) */
  h_display-=1;
  OUT_CRTC(0x01, h_display);

  /* Horizontal Blanking Start [0..255] (2044) */
  h_blanking_start-=1;
  OUT_CRTC(0x02, h_blanking_start);

  /* Horizontal Blanking End (Comparison) [0..63] (*/
  h_blanking_end-=1;
  OUT_CRTC(0x03, (IN_CRTC(0x03) & ~0x1f) | (h_blanking_end & 0x1f));
  OUT_CRTC(0x05, (IN_CRTC(0x05) & ~0x80) | ((h_blanking_end & 0x20) << 2));

  /* Horizontal Sync Start [0..255] (2044) */
  h_start-=1;
  OUT_CRTC(0x04, h_start);

  /* Horizontal Sync End (Comparison) [0..31] (248) */
  h_end-=1;
  OUT_CRTC(0x05, (IN_CRTC(0x05) & ~0x1f) | (h_end & 0x1f));


  /* Vertical Total [0..1023] */
  v_total-=2;
  OUT_CRTC(0x06, v_total);
  OUT_CRTC(0x07, (IN_CRTC(0x07) & ~0x21) | ((v_total & 0x100) >> 8) |
                                           ((v_total & 0x200) >> 4));

  /* Vertical Display End [0..1023] */
  v_display-=1;
  OUT_CRTC(0x12, v_display);
  OUT_CRTC(0x07, (IN_CRTC(0x07) & ~0x42) | ((v_display & 0x100) >> 7) |
                                           ((v_display & 0x200) >> 3));

  /* Vertical Blanking Start [0..1023] */
  v_blanking_start-=1;
  OUT_CRTC(0x15, v_blanking_start);
  OUT_CRTC(0x07, (IN_CRTC(0x07) & ~0x08) | ((v_blanking_start & 0x100) >> 5));
  OUT_CRTC(0x09, (IN_CRTC(0x09) & ~0x20) | ((v_blanking_start & 0x200) >> 4));

  /* Vertical Blanking End (Comparison) [0..255] */
  v_blanking_end-=1;
  OUT_CRTC(0x16, v_blanking_end);

  /* Vertical Sync Start [0..1023] */
  OUT_CRTC(0x10, v_start);
  OUT_CRTC(0x07, (IN_CRTC(0x07) & ~0x84) | ((v_start & 0x100) >> 6) |
                                           ((v_start & 0x200) >> 2));

  /* Vertical Sync End (Comparison) [0..31] */
  OUT_CRTC(0x11, (IN_CRTC(0x11) & ~0x0f) | (v_end & 0x0f));
}

static void std_get_timing(int *h_display,int *h_start,int *h_end,int *h_total,
                           int *v_display,int *v_start,int *v_end,int *v_total)
{
  int v_multiply = (IN_CRTC(0x17) & 0x04) ? 2 : 1;
  int h_sync, v_sync;

  *h_total=(IN_CRTC(0x00)+5) * 8;
  *h_display=(IN_CRTC(0x01)+1) * 8;

  h_sync=IN_CRTC(0x04);
  *h_start=(h_sync+1) * 8;
  *h_end=(IN_CRTC(0x05) & 0x1f);
  *h_end=((h_sync & ~0x1f)|(*h_end & 0x1f))+
          (((h_sync & 0x1f) > *h_end)?0x20:0x00);
  *h_end=((*h_end)+1) * 8;
  
  *v_total=((IN_CRTC(0x06) | ((IN_CRTC(0x07) & 0x01) << 8) |
            ((IN_CRTC(0x07) & 0x20) << 4))+2)
            * v_multiply;

  *v_display=((IN_CRTC(0x12) | ((IN_CRTC(0x07) & 0x02) << 7) |
              ((IN_CRTC(0x07) & 0x40) << 3))+1)
              * v_multiply;

  v_sync=(IN_CRTC(0x10) | ((IN_CRTC(0x07) & 0x04) << 6) |
          ((IN_CRTC(0x07) & 0x80) << 2));
  *v_start=v_sync * v_multiply;
  *v_end=(IN_CRTC(0x11) & 0xf);
  *v_end=((v_sync & ~0xf)|(*v_end & 0xf))+
          (((v_sync & 0xf) > *v_end)?0x10:0x00);
  *v_end=*v_end * v_multiply;

  /* Perform checkups */
  if(*v_start-*v_display > 512)
    *v_start-=512;
  if(*v_end-*v_display > 512)
    *v_end-=512;
  if(*v_total-*v_display > 512)
    *v_total-=512;
}

/* VGA Flags */

static void std_set_flags(int flags, int enable)
{
  if(flags & VGA_DOUBLESCAN)
    OUT_CRTC(0x09, (IN_CRTC(0x09) & ~0x80) | (enable ? 0x80:0));

  if(flags & (VGA_NHSYNC|VGA_NVSYNC)) {
    int reg=std_inmisc();

    if(flags & VGA_NHSYNC)
      reg=(reg & ~0x40) | (enable ? 0x40:0);
    if(flags & VGA_NVSYNC)
      reg=(reg & ~0x80) | (enable ? 0x80:0);
    std_outmisc(reg);
  }
}

static int std_get_flags()
{
  int flags=0, reg=std_inmisc();

  if(IN_CRTC(0x09) & 0x80)
    flags |= VGA_DOUBLESCAN;
  if(reg & 0x40)
    flags |= VGA_NHSYNC;
  if(reg & 0x80)
    flags |= VGA_NVSYNC;

  return flags;
}

static void std_wait_for_retrace()
{
  int i;

  /* wait during vertical retrace */
  for(i=0;(inb(crtc_index+0x0a) & 0x08) && i < VGA_TIMEOUT;i++);

  /* wait during active display */
  for(i=0;!(inb(crtc_index+0x0a) & 0x08) && i < VGA_TIMEOUT;i++);
}

static void std_set_error() {
  if(vga_verbose)
    printf("VGA Error: Function not supported with this video card!\n");
}

static void std_get_graphics_cursor_position(int *x, int *y) {
  std_set_error();
}

void change_std()
{
  iopl(3);

  /* Determine CRTC I/O Address */
  crtc_index=(std_inmisc() & 0x01)?0x3d0:0x3b0;

  vga_inmisc = std_inmisc;
  vga_incrtc = std_incrtc;
  vga_inseq  = std_inseq;
  vga_ingra  = std_ingra;
  vga_inattr = std_inattr;

  vga_outmisc = std_outmisc;
  vga_outcrtc = std_outcrtc;
  vga_outseq  = std_outseq;
  vga_outgra  = std_outgra;
  vga_outattr = std_outattr;

  vga_set_palette = std_set_palette;
  vga_set_palette_rgb = std_set_palette_rgb;
  vga_set_palette_vector = std_set_palette_vector;

  vga_get_palette_rgb = std_get_palette_rgb;
  vga_get_palette_vector = std_get_palette_vector;

  vga_set_dac_entry = std_set_dac_entry;
  vga_get_dac_entry = std_get_dac_entry;

  vga_set_text_cursor_size     = std_set_text_cursor_size;
  vga_set_text_cursor_location = std_set_text_cursor_location;
  vga_set_text_cell_height     = std_set_text_cell_height;
  vga_set_text_cell_width      = std_set_text_cell_width;
  vga_set_text_underline       = std_set_text_underline;
  vga_set_text_attr            = std_set_text_attr;

  vga_get_text_cursor_size     = std_get_text_cursor_size;
  vga_get_text_cursor_location = std_get_text_cursor_location;
  vga_get_text_cell_height     = std_get_text_cell_height;
  vga_get_text_cell_width      = std_get_text_cell_width;
  vga_get_text_underline       = std_get_text_underline;
  vga_get_text_attr            = std_get_text_attr;

  vga_set_display_start = std_set_display_start;
  vga_set_display_width = std_set_display_width;

  vga_get_display_start = std_get_display_start;
  vga_get_display_width = std_get_display_width;

  vga_set_clock = std_set_clock;
  vga_get_clock = std_get_clock;

  vga_set_timing = std_set_timing;
  vga_get_timing = std_get_timing;

  vga_set_flags = std_set_flags;
  vga_get_flags = std_get_flags;

  vga_graphics_cursor_enable = std_set_error;
  vga_graphics_cursor_select = std_set_error;

  vga_set_graphics_cursor_position = std_set_error;
  vga_get_graphics_cursor_position = std_get_graphics_cursor_position;

  vga_blank_screen = std_blank_screen;
  vga_wait_for_retrace = std_wait_for_retrace;
}
