/* vga.c */

#include "vga.h"

int vga_chipset;	// Chipset Type (0=Standard VGA)
int vga_clockchip;	// Clockchip Type (0=Super VGA)
int vga_memory;		// Total onboard memory in kB
int vga_maxclock;	// Maximum pixel clock in kHz

int vga_verbose;	// 1=Report all actions
int vga_num_cards;      // Number of VGA cards detected

int vga_options;	// Options set by the user program

static struct pcivendor {
  struct pcimodel *model;
  int (*init)(int card, int index);
} vendors[]={
  {model_svga, std_init},
  {model_nvidia, nv_init},
  {model_ati, ati_init},
  {0}
};


// Initialize the VGA Library (Primary VGA Card).
//
int vga_init()
{
  return vga_init_card(1, 0);
}

int vga_init_card(int card, int init)
{
  int i, j=0;

  if(!vga_num_cards)
    vga_num_cards=pci_find_vendor_vga(vga_verbose);

  card--;
  if(card < 0 || card >= vga_num_cards) {
    printf("VGA Card #%d not found.\n", card+1);
    return 0;
  }

  if(vga_verbose)
    printf("Initializing VGA Card #%d\n", card+1);

  /* Scan for valid VGA vendor */
  for(i=0;vendors[i].model;i++) {
    for(j=0;vendors[i].model[j].vendor;j++)
      if((card_list[card].vendor & 0xffff) == vendors[i].model[j].vendor &&
         (card_list[card].vendor >> 16) >= vendors[i].model[j].lo &&
         ((vendors[i].model[j].hi < vendors[i].model[j].lo &&
           (card_list[card].vendor >> 16) == vendors[i].model[j].lo) ||
          (card_list[card].vendor >> 16) <= vendors[i].model[j].hi))
        break;
    if(vendors[i].model[j].vendor)
      break;
  }

  if(!pci_enable_device(card_list[card].device) || init) {
    v86bios_init(card_list[card].device);
    usleep(100000);
  }

  if(vendors[i].model)
    return vendors[i].init(card, j);

  return std_init(card, -1);
}


/* Function Definitions */


/* VGA lowlevel registers */
int  (*vga_inmisc)(void);
int  (*vga_incrtc)(int i);
int  (*vga_inseq)(int i);
int  (*vga_ingra)(int i);
int  (*vga_inattr)(int i);
void (*vga_outmisc)(int val);
void (*vga_outcrtc)(int i, int val);
void (*vga_outseq)(int i, int val);
void (*vga_outgra)(int i, int val);
void (*vga_outattr)(int i, int val);


/* VGA textmode parameters */
void (*vga_set_text_cursor_size)(int start, int end);
void (*vga_set_text_cursor_location)(int address);
void (*vga_set_text_cell_height)(int height);
void (*vga_set_text_cell_width)(int width);
void (*vga_set_text_underline)(int scanline);
void (*vga_set_text_attr)(int flags, int enable);

void (*vga_get_text_cursor_size)(int *start, int *end);
int  (*vga_get_text_cursor_location)(void);
int  (*vga_get_text_cell_height)(void);
int  (*vga_get_text_cell_width)(void);
int  (*vga_get_text_underline)(void);
int  (*vga_get_text_attr)(void);

/* VGA text/graphics parameters */
void (*vga_set_display_start)(int address);
void (*vga_set_display_width)(int width);
void (*vga_set_timing)(int h_display,int h_start,int h_end,int h_total,
                       int v_display,int v_start,int v_end,int v_total);
void (*vga_set_flags)(int flags, int enable);
int  (*vga_set_clock)(int khz);
void (*vga_set_palette)(int index, int r, int g, int b);
void (*vga_set_palette_rgb)(int index, int rgb);
void (*vga_set_palette_vector)(int index, int num,
                               const unsigned char *vector);
void (*vga_set_dac_entry)(int reg, int index);

int  (*vga_get_display_start)(void);
int  (*vga_get_display_width)(void);
void (*vga_get_timing)(int*h_display,int*h_start,int*h_end,int*h_total,
                       int*v_display,int*v_start,int*v_end,int*v_total);
int  (*vga_get_flags)(void);
int  (*vga_get_clock)(void);
int  (*vga_get_palette_rgb)(int index);
void (*vga_get_palette_vector)(int index,int num,unsigned char *vector);
int  (*vga_get_dac_entry)(int reg);

void (*vga_graphics_cursor_enable)(int enable);
void (*vga_graphics_cursor_select)(int index);
void (*vga_set_graphics_cursor_position)(int x, int y);
void (*vga_get_graphics_cursor_position)(int *x, int *y);

/* VGA control */
void (*vga_blank_screen)(int blank);
void (*vga_wait_for_retrace)(void);
