/* vga.h */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/io.h>

#include "pci.h"

struct pcimodel {
  unsigned short vendor;
  unsigned short lo;
  unsigned short hi;
  int rev;
  char *name;
  int clockchip;
};

extern struct pcimodel model_svga[], model_nvidia[], model_ati[];


/* VGA Flags */
#define VGA_NHSYNC	0x1
#define VGA_NVSYNC	0x2
#define VGA_DOUBLESCAN	0x4
#define VGA_INTERLACE	0x8
#define VGA_CSYNC	0x10

/* VGA Extended Text Attributes */
#define VGA_TEXT_BLINK      0x1  // 1=Blink enable, 0=8-bit background
#define VGA_TEXT_FONT_512   0x2  // 1=Hilite enable, 0=512-char font
#define VGA_TEXT_8BIT_DAC   0x4  // 1=DAC uses 8-bit color values for palette
#define VGA_TEXT_DUP_PIXEL9 0x8  // 1=Font cells C0-DF have pixel 9 duplicated

#define VGA_TIMEOUT	200000

extern int std_init(int card, int index);
extern int nv_init(int card, int index);
extern int ati_init(int card, int index);

enum chipsets {
  VGA_CS_SVGA, VGA_CS_NVIDIA, VGA_CS_ATI, VGA_CS_MATROX, VGA_CS_S3,
  VGA_CS_CIRRUS, VGA_CS_ET3000, VGA_CS_ET4000, VGA_CS_ET6000, VGA_CS_TVGA8900,
  VGA_CS_TVGA9000, VGA_CS_PVGA1, VGA_CS_WDC90C0X, VGA_CS_WDC90C1X,
  VGA_CS_WDC90C2X, VGA_CS_WDC90C3X, VGA_CS_ATI28800, VGA_CS_ATIMACH32,
  VGA_CS_ATIMACH64, VGA_CS_HEADLAND, VGA_CS_ALI, VGA_CS_ALI2101, VGA_CS_OTI67,
  VGA_CS_OTI77, VGA_CS_OTI87, VGA_CS_SIS, VGA_CS_REALTEK, VGA_CS_ARK,
  VGA_CS_NCR22E, VGA_CS_NCR32, VGA_CS_GVGA, VGA_CS_MX, VGA_CS_TGUI,
  VGA_CS_NEOMAGIC,
};

enum clockchips {
  VGA_CLOCK_ICD2061A, VGA_CLOCK_ICS9161A, VGA_CLOCK_ICS2595, VGA_CLOCK_DCS2834,
  VGA_CLOCK_ICS5300, VGA_CLOCK_ICS5341, VGA_CLOCK_ICS5342, VGA_CLOCK_SC11412,
  VGA_CLOCK_S3GENDAC, VGA_CLOCK_S3_SDAC, VGA_CLOCK_S3TRIO, VGA_CLOCK_S3VIRGE,
  VGA_CLOCK_TI3025, VGA_CLOCK_TI3026, VGA_CLOCK_CH8391, VGA_CLOCK_STG1703,
  VGA_CLOCK_IBMRGB5XX, VGA_CLOCK_ET6000, VGA_CLOCK_MACH64, VGA_CLOCK_TGUI9320,
  VGA_CLOCK_TGUI9440, VGA_CLOCK_CYBER938X, VGA_CLOCK_ATT20C408,
  VGA_CLOCK_MGA1064, VGA_CLOCK_LAGUNA,
};

/* VGA Options (Flags) */
#define VGA_OPT_SLOW_DRAM	0x10   // Memory speeds for Cirrus Logic & S3
#define VGA_OPT_MED_DRAM	0x20
#define VGA_OPT_FAST_DRAM	0x40
#define VGA_OPT_XFAST_DRAM	0x80
#define VGA_OPT_HIBIT_HIGH	0x100  // ET4000-specific
#define VGA_OPT_SWAP_HIBIT	0x200  // WDC-specific
#define VGA_OPT_LEGEND		0x400  // Set on Sigma Legend cards
#define VGA_OPT_SPEA_MERCURY	0x800  // Another ET4000-specific
#define VGA_OPT_S3_HSTEXT	0x1000 // S3 HighSpeed > 36MHz Pixel Clocks
#define VGA_OPT_CLOCKCHIP_X	0x2000 // (Not sure what this does yet)
#define VGA_OPT_TOPLEFT		0x4000 // Neomagic-specific
#define VGA_OPT_STRETCH		0x8000 // Neomagic-specific

extern int vga_chipset;		// Chipset type (one of the above)
extern int vga_clockchip;	// Clockchip type
extern int vga_memory;		// Total onboard memory in kB
extern int vga_maxclock;	// Maximum pixel clock in kHz

extern int vga_verbose;

extern int vga_options;		// Options set by the user program

/* Macros */

#define MIN(a,b)  ((a) < (b) ? (a) : (b))
#define MAX(a,b)  ((a) > (b) ? (a) : (b))


/* Function Declarations */

/* Initialize the VGA Library */
extern int  vga_init(void);
extern int  vga_init_device(int device);
extern int  vga_init_card(int card, int init);

extern int  vga_switch_card(int card);
extern void change_std(void);

/* VGA lowlevel registers */
extern int  (*vga_inmisc)(void);
extern int  (*vga_incrtc)(int i);
extern int  (*vga_inseq)(int i);
extern int  (*vga_ingra)(int i);
extern int  (*vga_inattr)(int i);
extern void (*vga_outmisc)(int val);
extern void (*vga_outcrtc)(int i, int val);
extern void (*vga_outseq)(int i, int val);
extern void (*vga_outgra)(int i, int val);
extern void (*vga_outattr)(int i, int val);


/* VGA textmode parameters */
extern void (*vga_set_text_cursor_size)(int start, int end);
extern void (*vga_set_text_cursor_location)(int address);
extern void (*vga_set_text_cell_height)(int height);
extern void (*vga_set_text_cell_width)(int width);
extern void (*vga_set_text_underline)(int scanline);
extern void (*vga_set_text_attr)(int flags, int enable);

extern void (*vga_get_text_cursor_size)(int *start, int *end);
extern int  (*vga_get_text_cursor_location)(void);
extern int  (*vga_get_text_cell_height)(void);
extern int  (*vga_get_text_cell_width)(void);
extern int  (*vga_get_text_underline)(void);
extern int  (*vga_get_text_attr)(void);

/* VGA text/graphics parameters */
extern void (*vga_set_display_start)(int address);
extern void (*vga_set_display_width)(int width);
extern void (*vga_set_timing)(int h_display,int h_start,int h_end,int h_total,
                              int v_display,int v_start,int v_end,int v_total);
extern void (*vga_set_flags)(int flags, int enable);
extern int  (*vga_set_clock)(int khz);
extern void (*vga_set_palette)(int index, int r, int g, int b);
extern void (*vga_set_palette_rgb)(int index, int rgb);
extern void (*vga_set_palette_vector)(int index, int num,
                                      const unsigned char *vector);
extern void (*vga_set_dac_entry)(int reg, int index);

extern int  (*vga_get_display_start)(void);
extern int  (*vga_get_display_width)(void);
extern void (*vga_get_timing)(int*h_display,int*h_start,int*h_end,int*h_total,
                              int*v_display,int*v_start,int*v_end,int*v_total);
extern int  (*vga_get_flags)(void);
extern int  (*vga_get_clock)(void);
extern int  (*vga_get_palette_rgb)(int index);
extern void (*vga_get_palette_vector)(int index,int num,unsigned char *vector);
extern int  (*vga_get_dac_entry)(int reg);

extern void (*vga_graphics_cursor_enable)(int enable);
extern void (*vga_graphics_cursor_select)(int index);
extern void (*vga_set_graphics_cursor_position)(int x, int y);
extern void (*vga_get_graphics_cursor_position)(int *x, int *y);

/* VGA control */
extern void (*vga_blank_screen)(int blank);
extern void (*vga_wait_for_retrace)(void);

/* V86Bios */
extern void v86bios_init(unsigned short slot);

/* clock.c */
extern int  std_set_clock(int freq);
extern int  std_get_clock(void);
